@@ -34,7 +34,7 @@ is the one that supports the extraction of the source package.
 use strict;
 use warnings;
 
-our $VERSION = '2.01';
+our $VERSION = '2.02';
 our @EXPORT_OK = qw(
     get_default_diff_ignore_regex
     set_default_diff_ignore_regex
@@ -44,11 +44,10 @@ our @EXPORT_OK = qw(
 use Exporter qw(import);
 use POSIX qw(:errno_h :sys_wait_h);
 use Carp;
-use Cwd qw(realpath);
 use File::Temp;
-use File::Find;
 use File::Copy qw(cp);
 use File::Basename;
+use File::Spec;
 
 use Dpkg::Gettext;
 use Dpkg::ErrorHandling;
@@ -56,10 +55,11 @@ use Dpkg::Control;
 use Dpkg::Checksums;
 use Dpkg::Version;
 use Dpkg::Compression;
-use Dpkg::Path qw(check_files_are_the_same);
+use Dpkg::Path qw(check_files_are_the_same check_directory_traversal);
 use Dpkg::Vendor qw(run_vendor_hook);
 use Dpkg::Source::Format;
 use Dpkg::OpenPGP;
+use Dpkg::OpenPGP::ErrorCodes;
 
 my $diff_ignore_default_regex = '
 # Ignore general backup files
@@ -214,6 +214,7 @@ sub new {
         format => Dpkg::Source::Format->new(),
         options => {},
         checksums => Dpkg::Checksums->new(),
+        openpgp => Dpkg::OpenPGP->new(needs => { api => 'verify' }),
     };
     bless $self, $class;
     if (exists $args{options}) {
@@ -227,6 +228,13 @@ sub new {
         $self->upgrade_object_type(0);
         $self->init_options();
     }
+
+    if ($self->{options}{require_valid_signature}) {
+        $self->{report_verify} = \&error;
+    } else {
+        $self->{report_verify} = \&warning;
+    }
+
     return $self;
 }
 
@@ -248,9 +256,16 @@ sub init_options {
          'debian/source/local-patch-header',
          'debian/files',
          'debian/files.new';
+    $self->{options}{copy_orig_tarballs} //= 0;
+
     # Skip debianization while specific to some formats has an impact
     # on code common to all formats
     $self->{options}{skip_debianization} //= 0;
+    $self->{options}{skip_patches} //= 0;
+
+    # Set default validation checks.
+    $self->{options}{require_valid_signature} //= 0;
+    $self->{options}{require_strong_checksums} //= 0;
 
     # Set default compressor for new formats.
     $self->{options}{compression} //= 'xz';
@@ -318,7 +333,7 @@ Returns the filename of the DSC file.
 
 sub get_filename {
     my $self = shift;
-    return $self->{basedir} . $self->{filename};
+    return File::Spec->catfile($self->{basedir}, $self->{filename});
 }
 
 =item $p->get_files()
@@ -359,7 +374,8 @@ sub check_checksums {
                 $warn_on_weak = 1;
             }
         }
-	$checksums->add_from_file($self->{basedir} . $file, key => $file);
+	my $pathname = File::Spec->catfile($self->{basedir}, $file);
+	$checksums->add_from_file($pathname, key => $file);
     }
 
     warning(g_('source package uses only weak checksums')) if $warn_on_weak;
@@ -387,7 +403,7 @@ sub find_original_tarballs {
     foreach my $dir ('.', $self->{basedir}, $self->{options}{origtardir}) {
         next unless defined($dir) and -d $dir;
         opendir(my $dir_dh, $dir) or syserr(g_('cannot opendir %s'), $dir);
-        push @tar, map { "$dir/$_" } grep {
+        push @tar, map { File::Spec->catfile($dir, $_) } grep {
 		($opts{include_main} and
 		 /^\Q$basename\E\.orig\.tar\.$opts{extension}$/) or
 		($opts{include_supplementary} and
@@ -410,6 +426,24 @@ sub get_upstream_signing_key {
     return "$dir/debian/upstream/signing-key.asc";
 }
 
+=item $p->armor_original_tarball_signature($bin, $asc)
+
+Convert a signature from binary to ASCII armored form. If the signature file
+does not exist, it is a no-op. If the signature file is already ASCII armored
+then simply copy it, otherwise convert it from binary to ASCII armored form.
+
+=cut
+
+sub armor_original_tarball_signature {
+    my ($self, $bin, $asc) = @_;
+
+    if (-e $bin) {
+        return $self->{openpgp}->armor('SIGNATURE', $bin, $asc);
+    }
+
+    return;
+}
+
 =item $p->check_original_tarball_signature($dir, @asc)
 
 Verify the original upstream tarball signatures @asc using the upstream
@@ -428,21 +462,15 @@ sub check_original_tarball_signature {
         return;
     }
 
-    my $keyring = File::Temp->new(UNLINK => 1, SUFFIX => '.gpg');
-    my %opts = (
-        require_valid_signature => $self->{options}{require_valid_signature},
-    );
-    Dpkg::OpenPGP::import_key($upstream_key,
-        %opts,
-        keyring => $keyring,
-    );
-
     foreach my $asc (@asc) {
-        Dpkg::OpenPGP::verify_signature($asc,
-            %opts,
-            keyrings => [ $keyring ],
-            datafile => $asc =~ s/\.asc$//r,
-        );
+        my $datafile = $asc =~ s/\.asc$//r;
+
+        info(g_('verifying %s'), $asc);
+        my $rc = $self->{openpgp}->verify($datafile, $asc, $upstream_key);
+        if ($rc) {
+            $self->{report_verify}->(g_('cannot verify upstream tarball signature for %s: %s'),
+                                     $datafile, openpgp_errorcode_to_string($rc));
+        }
     }
 }
 
@@ -471,22 +499,21 @@ then any problem will result in a fatal error.
 sub check_signature {
     my $self = shift;
     my $dsc = $self->get_filename();
-    my @keyrings;
+    my @certs;
+
+    push @certs, $self->{openpgp}->get_trusted_keyrings();
 
-    if (length $ENV{HOME} and -r "$ENV{HOME}/.gnupg/trustedkeys.gpg") {
-        push @keyrings, "$ENV{HOME}/.gnupg/trustedkeys.gpg";
-    }
     foreach my $vendor_keyring (run_vendor_hook('package-keyrings')) {
         if (-r $vendor_keyring) {
-            push @keyrings, $vendor_keyring;
+            push @certs, $vendor_keyring;
         }
     }
 
-    my %opts = (
-        keyrings => \@keyrings,
-        require_valid_signature => $self->{options}{require_valid_signature},
-    );
-    Dpkg::OpenPGP::verify_signature($dsc, %opts);
+    my $rc = $self->{openpgp}->inline_verify($dsc, undef, @certs);
+    if ($rc) {
+        $self->{report_verify}->(g_('cannot verify inline signature for %s: %s'),
+                                 $dsc, openpgp_errorcode_to_string($rc));
+    }
 }
 
 sub describe_cmdline_options {
@@ -539,7 +566,7 @@ sub extract {
             my $dst = File::Spec->catfile($destdir, $orig);
             if (not check_files_are_the_same($src, $dst, 1)) {
                 cp($src, $dst)
-                    or syserror(g_('cannot copy %s to %s'), $src, $dst);
+                    or syserr(g_('cannot copy %s to %s'), $src, $dst);
             }
         }
     }
@@ -548,22 +575,10 @@ sub extract {
     $self->do_extract($newdirectory);
 
     # Check for directory traversals.
-    if (not $self->{options}{skip_debianization}) {
-        my $canon_newdir = realpath($newdirectory);
-        my $check_symlinks = sub {
-            my $canon_pathname = realpath($_);
-            return if $canon_pathname =~ m/^\Q$canon_newdir\E/;
-
-            error(g_("pathname '%s' points outside source root"), $_);
-        };
+    if (not $self->{options}{skip_debianization} and not $self->{no_check}) {
         # We need to add a trailing slash to handle the debian directory
         # possibly being a symlink.
-        find({
-            wanted => $check_symlinks,
-            no_chdir => 1,
-            follow => 1,
-            follow_skip => 2,
-        }, "$newdirectory/debian/");
+        check_directory_traversal($newdirectory, "$newdirectory/debian/");
     }
 
     # Store format if non-standard so that next build keeps the same format
@@ -688,6 +703,10 @@ sub write_dsc {
 
 =head1 CHANGES
 
+=head2 Version 2.02 (dpkg 1.21.10)
+
+New method: armor_original_tarball_signature().
+
 =head2 Version 2.01 (dpkg 1.20.1)
 
 New method: get_upstream_signing_key().
